"""Interface for default modules time and datetime"""

# Standard modules
import time as _time
import datetime as _datetime


def wait(duration: float) -> None:
    """Wait for a specific amount of seconds"""
    _time.sleep(duration)


def epoch_start() -> _time.struct_time:
    """Get the datetime the epoch started"""
    return _time.gmtime(0)


def epoch_time() -> float:
    """Get the time in seconds since the epoch (Get epoch start with epoch_start())"""
    return _time.time()


def epoch_time_ns() -> int:
    """Get the time in nanoseconds since the epoch (Get epoch start with epoch_start())"""
    return _time.time_ns()


def bench_time() -> float:
    """Get a time in seconds for calculating a duration by getting the difference"""
    return _time.perf_counter()


def bench_time_ns() -> int:
    """Get a time in nanoseconds for calculating a duration by getting the difference"""
    return _time.perf_counter_ns()


def run_time() -> float:
    """Get the time in seconds since the process started"""
    return _time.process_time()


def run_time_ns() -> int:
    """Get the time in nanoseconds since the process started"""
    return _time.process_time_ns()


def time_f_hms() -> str:
    """Get the time as a string in the format H:M:S"""
    return _datetime.datetime.now().strftime("%H:%M:%S")


def date_f_dmy() -> str:
    """Get the date as a string in the format D.M.Y"""
    return _datetime.datetime.now().strftime("%d.%b:%y")


def datetime_f_dmy_hms() -> str:
    """Get the datetime as a string in the format D.M.Y H:M:S"""
    return _datetime.datetime.now().strftime("%d.%b.%y %H:%M:%S")


def datetime_f_ymd_hms() -> str:
    """Get the datetime as a string in the format Y-M-D_Hh-Mm-Ss"""
    return _datetime.datetime.now().strftime("%Y-%m-%d_%Hh-%Mm-%Ss")


def datetime_f(datetime_format: str) -> str:
    """Get the datetime as a string in a custom format"""
    return _datetime.datetime.now().strftime(datetime_format)



"""Contains the 2D vector class"""

# Standard modules
from typing import Union, AnyStr, Tuple, Self, List, Dict
import math


class Vec(tuple):
    """2D vector class"""

    def __new__(cls, x: Union[int, float], y: Union[int, float]) -> Self:
        return tuple.__new__(cls, (x, y))

    @property
    def x(self) -> Union[int, float]:
        return self[0]

    @property
    def y(self) -> Union[int, float]:
        return self[1]

    @property
    def xy(self) -> Self:
        return self

    @property
    def yx(self) -> Self:
        return Vec(self[1], self[0])

    @property
    def xx(self) -> Self:
        return Vec(self[0], self[0])

    @property
    def yy(self) -> Self:
        return Vec(self[1], self[1])

    @property
    def rounded(self) -> bool:
        return isinstance(self[0], int) and isinstance(self[1], int)

    @property
    def normalized(self) -> bool:
        return self.length == 1

    @property
    def length(self) -> float:
        return math.sqrt(self[0] ** 2 + self[1] ** 2)

    @property
    def length_squared(self) -> float:
        return self[0] ** 2 + self[1] ** 2

    def list(self) -> List[Union[int, float]]:
        """Create a list with the x and y values as elements"""
        return [*self]

    def dict(self, keys: Tuple[AnyStr, AnyStr] = ("x", "y")) -> Dict[AnyStr, Union[int, float]]:
        """Create a dictionary with the x and y values as value of specific keys"""
        return {keys[0]: self[0], keys[1]: self[1]}

    def distance(self, other: Self) -> float:
        """Calculate the distance to another vector"""
        return math.sqrt(pow(self[0] - other[0], 2) + pow(self[1] - other[1], 2))

    def distance_squared(self, other: Self) -> float:
        """Calculate the squared distance to another vector"""
        return abs(self[0] - other[0]) + abs(self[1] - other[1])

    def normalize(self) -> Self:
        """Normalize the vector to a length of 1"""
        if self.length == 0:
            return Vec(0, 0)
        return self.__truediv__(self.length)

    def round(self, n: int = None) -> Self:
        """Round the vector to a vector with integer values"""
        return self.__round__(n)

    def floor(self) -> Self:
        """Floor the vector to a vector with integer values"""
        return self.__floor__()

    def ceil(self) -> Self:
        """Ceil the vector to a vector with integer values"""
        return self.__ceil__()

    def __eq__(self, other: Self) -> bool:
        return self[0] == other[0] and self[1] == other[1]

    def __ne__(self, other: Self) -> bool:
        return self[0] != other[0] or self[1] != other[1]

    def __add__(self, other: Self) -> Self:
        return Vec(self[0] + other[0], self[1] + other[1])

    def __sub__(self, other: Self) -> Self:
        return Vec(self[0] - other[0], self[1] - other[1])

    def __mul__(self, factor: Union[int, float]) -> Self:
        return Vec(self[0] * factor, self[1] * factor)

    def __truediv__(self, divisor: Union[int, float]) -> Self:
        return Vec(self[0] / divisor, self[1] / divisor)

    def __floordiv__(self, divisor: Union[int, float]) -> Self:
        return Vec(self[0] // divisor, self[1] // divisor)

    def __pow__(self, power: float, modulo: float = None):
        return Vec(pow(self[0], power, modulo), pow(self[1], power, modulo))

    def __round__(self, n: int = None) -> Self:
        return Vec(round(self[0], n), round(self[1], n))

    def __floor__(self) -> Self:
        return Vec(math.floor(self[0]), math.floor(self[1]))

    def __ceil__(self) -> Self:
        return Vec(math.ceil(self[0]), math.ceil(self[1]))

    def __abs__(self) -> Self:
        return Vec(abs(self[0]), abs(self[1]))

    def __pos__(self) -> Self:
        return self

    def __neg__(self) -> Self:
        return Vec(-self[0], -self[1])

    def __repr__(self) -> AnyStr:
        return f"Vec[id={id(self)}, x={self[0]}, y={self[1]}]"

    def __str__(self) -> AnyStr:
        return f"({self[0]}, {self[1]})"



"""Contains the clock class for time handling in pygame"""

# Standard modules
from typing import AnyStr, List

# External modules
import pygame as pg

# Local modules
from ..utils import time


class Clock:
    """Clock class for time handling in pygame"""

    def __init__(self, max_fps: int, statistic_speed: int = 5) -> None:

        self.max_fps: int = max_fps
        self.limiter: pg.time.Clock = pg.time.Clock()

        self.start_time: float = 0
        self.frame_count: int = 0
        self.frame_start: float = 0
        self.frame_end: float = 0
        self.frame_durations: List[float] = [0] * statistic_speed
        self.frame_duration: float = 1
        self.last_update: float = 0

        self.delta_time: float = 0
        self.last_time: float = 0

        self.speed_factor: float = 1

    @property
    def real_fps(self) -> float:
        return round(self.limiter.get_fps(), 5)

    @property
    def available_fps(self) -> float:
        return round(1 / self.frame_duration, 5)

    def tick(self, max_fps: int = None) -> None:
        """Next frame (Calculates statistics and delta time and waits until the next frame should be calculated)"""

        self.frame_end: float = time.bench_time()

        self.frame_durations.append(self.frame_end - self.frame_start)
        del self.frame_durations[0]
        if time.bench_time() - self.last_update > 0.3:
            self.frame_duration: float = sum(self.frame_durations) / len(self.frame_durations)
            if self.frame_duration == 0:
                self.frame_duration: float = 1
            self.last_update: float = time.bench_time()

        if max_fps is not None:
            self.max_fps: int = max_fps

        self.limiter.tick(self.max_fps)

        self.frame_count += 1

        self.delta_time: float = time.bench_time() - self.last_time
        self.last_time: float = time.bench_time()
        self.delta_time *= self.max_fps * self.speed_factor

        self.frame_start: float = time.bench_time()

    def __repr__(self) -> AnyStr:
        return f"Clock[id={id(self)}, real_fps={self.real_fps}, available_fps={self.available_fps}, max_fps={self.max_fps}, speed_factor={self.speed_factor}, frame_count={self.frame_count}]"



"""Nikocraft Python Library - Window"""

# Modules
from . import rgb

# Classes
from .window import Window
from .vector2d import Vec
from .vector3d import Vec3
from .clock import Clock

__all__ = [
    "rgb",
    "Window",
    "Vec",
    "Vec3",
    "Clock"
]



"""Contains the main class of the application"""

# Standard modules
from abc import ABC, abstractmethod
from typing import AnyStr
import os
import sys
import ctypes

# External modules


# Local modules
from .constants import *


class App(ABC):
    """Main class of the application"""

    _initialized = False

    def __init__(self, args: list[AnyStr] = None) -> None:

        self.args: list[AnyStr] = args if args else sys.argv
        self.debug: bool = "-d" in args or "--debug" in args

        self.exit_code: int = 0

        self._initialized: bool = True

    def start(self) -> int:
        """Start the application

        Returns the exit code
        """

        assert self._initialized, "Application was not initialized!"

        self.run()

        self.quit()
        return self.exit_code

    @abstractmethod
    def run(self) -> None:
        """Application execution

        Called on starting -
        Don't call this method manually
        """

        pass

    @abstractmethod
    def quit(self) -> None:
        """Shutdown tasks

        Called before exiting -
        Don't call this method manually
        """

        pass

    @staticmethod
    def disable_resolution_scaling() -> None:
        """Disable resolution scaling on Windows

        Returns nothing
        """

        if os.name == "nt":
            ctypes.windll.user32.SetProcessDPIAware()



"""Nikocraft Python Library"""

# Environment variables
import os
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = ""
del os

# Modules
from . import window
from .utils import time
from .window import rgb

# Classes
from .app import App
from .utils.config import Config
from .window.vector2d import Vec
from .window.vector3d import Vec3

# Constants
from .constants import AUTHOR, VERSION

__all__ = [
    "window",
    "time",
    "rgb",
    "App",
    "Config",
    "Vec",
    "Vec3",
    "AUTHOR",
    "VERSION"
]



"""Contains the config class for saving JSON configurations"""

# Standard modules
from typing import AnyStr, Dict, Any
from abc import ABC, abstractmethod
import json

# External modules
import pygame as pg

# Local modules
from ..utils import time


class Config(ABC):
    """Config class for saving JSON configurations"""

    def __init__(self, path: str) -> None:

        self.path: str = path

    def load(self) -> None:
        """Load data from JSON file at path"""

        print(vars(self))

    @abstractmethod
    def load_attribute(self, attribute: str, data: Dict) -> Any:
        """Load an attribute from data (Return result or None for automatic loading)

        *Called on loading -
        Don't call this method manually*
        """

        pass

    def __repr__(self) -> AnyStr:
        return f"Config[id={id(self)}, path={self.path}]"



"""Setup Script"""

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="nikocraft",
    version="0.0.1",
    description="Nikocraft Python Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Nikocraft123/Python-Nikocraft",
    author="Nikocraft",
    author_email="nikocraft@gmx.net",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",

        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",

        "Topic :: Multimedia :: Graphics",
        "Topic :: Games/Entertainment",
        "Topic :: Utilities",
    ],
    keywords="python, pygame, utils, framework, development",
    packages=find_packages(),
    python_requires=">=3.11, <4",
    install_requires=["pygame"],
    extras_require={
        "cv": ["opencv-python"],
    },
    # package_data={
    #     "sample": ["package_data.dat"],
    # },
    project_urls={
        "Source": "https://github.com/Nikocraft123/Python-Nikocraft",
    },
)


"""Contains the window class for GUI management"""

# Standard modules
from abc import ABC, abstractmethod
import os
import ctypes

# External modules
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = ""
import pygame as pg

# Local modules
from ..constants import *
from ..app import App
from .vector2d import Vec
from .clock import Clock


class Window(ABC):
    """Window class for the GUI management"""

    _initialized = False

    def __init__(self, app: App) -> None:

        self.screen: pg.Surface = pg.Surface((0, 0))
        self.target_dimension: Vec = Vec(DEFAULT_WIDTH, DEFAULT_HEIGHT)
        self.running: bool = False
        self.clock: Clock = Clock(DEFAULT_FPS)
        self.flags: int = 0
        self.option_auto_update_screen: bool = True
        self.option_auto_quit: bool = True

        pg.init()

        self._initialized = True

    @property
    def width(self):
        return self.screen.get_width()

    @property
    def height(self):
        return self.screen.get_height()

    @property
    def dimension(self):
        return Vec(self.screen.get_width(), self.screen.get_height())

    def open(self) -> None:
        """Open the window

        *Returns nothing*
        """

        assert self._initialized, "Application was not initialized!"

        self.screen: pg.Surface = pg.display.set_mode(self.target_dimension, self.flags)
        self.init()

        self.running = True

        while self.running:

            self.clock.tick()

            for event in pg.event.get():

                self.event(event)

                if event.type == pg.QUIT and self.option_auto_quit:
                    self.running = False

            self.render()

            if self.option_auto_update_screen:
                pg.display.flip()

        self.quit()
        pg.quit()

    @abstractmethod
    def init(self) -> None:
        """Startup tasks

        *Called after window is opened -
        Don't call this method manually*
        """

        pass

    @abstractmethod
    def event(self, event: pg.event.Event) -> None:
        """Handle pygame event

        *Called when event is fired*
        """

        pass

    @abstractmethod
    def render(self) -> None:
        """Render screen

        *Called every frame*
        """

        pass

    @abstractmethod
    def quit(self) -> None:
        """Shutdown tasks

        *Called before closing -
        Don't call this method manually*
        """

        pass
